unit TabListBox;

interface

uses
  Windows, Messages, SysUtils, Classes, StdCtrls, Controls;
//  Forms, Dialogs, Graphics;

type
  ETabListBoxError = class(Exception);

  TTabListBox = class(TListBox)
  private
    { Private declarations }

    FlongestString: Word;
    FNumTabStops: Word;
    FTabStops: PWord;
    FSizeAfterDel: Boolean;
    function GetLBStringLength(S: String): Word;
    procedure FindLongestString;
    procedure SetScrollLength(S: String);
    procedure LBAddString(var Msg: TMessage); message lb_AddString;
    procedure LBInsertString(var Msg: TMessage); message lb_InsertString;
    procedure LBDeleteString(var Msg: Tmessage); message lb_DeleteString;
  protected
    { Protected declarations }

    procedure CreateParams(var Params: TCreateParams); override;
  public
    { Public declarations }

    constructor Create(AOwner: TComponent); override;
    procedure SetTabStops(A: array of Word);
  published
    { Published declarations }

    property SizeAfterDel: Boolean read FSizeAfterDel write FSizeAfterDel default True;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Nov', [TTabListBox]);
end;

constructor TTabListBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FSizeAfterDel:= true;
  FNumTabStops:= 1;
  GetMem(FTabStops, SizeOf(Word) * FNumTabStops);
  FTabStops^:= 8 * LoWord(GetDialogbaseUnits);
end;

procedure TTabListBox.SetTabStops(A: array of Word);
var
  i, TempTab: Word;
  TempBuf: PWord;
begin
  TempTab:= High(A) + 1;
  GetMem(TempBuf, SizeOf(A));
  Move(A, TempBuf^,SizeOf(A));
  for i:=0 to TempTab -1 do
    A[i]:= A[i] * 4 div LoWord(GetDialogBaseUnits);
  if Perform(lb_SetTabStops, TempTab, LongInt(@A)) = 0 then
  begin
    FreeMem(TempBuf, SizeOf(Word) * TempTab);
    raise ETabListBoxError.Create('Failed to set tabs.');
  end
  else begin
    FreeMem(FtabStops, SizeOf(Word) * FNumTabStops);
    FNumTabStops:= TempTab;
    FTabStops:= TempBuf;
    FindLongestString;
    Invalidate;
  end;
end;

procedure TTabListBox.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style:= Params.Style or lbs_UseTabStops or ws_HScroll;
end;

function TTabListBox.GetLBStringLength(S: String): Word;
var Size: Integer;
begin
  Canvas.Font:= Font;
  Result:= LoWord(GetTabbedTextExtent(Canvas.Handle, PChar(S), StrLen(PChar(S)), FNumTabStops, FTabStops^));
  Size:= Canvas.TextWidth('X');
  inc(Result, Size);
end;

procedure TTabListBox.SetScrollLength(S: String);
var Extent: Word;
begin
  Extent:= GetLBStringLength(S);
  if Extent > FLongestString then
  begin
    FLongestString:= Extent;
    Perform(lb_SetHorizontalExtent, Extent, 0);
  end;
end;

procedure TTabListBox.LBInsertString(var Msg: TMessage);
begin
  inherited;
  SetScrollLength(PChar(Msg.lParam));
end;

procedure TTabListBox.LBAddString(var Msg: TMessage);
begin
  inherited;
  SetScrollLength(PChar(Msg.lParam));
end;

procedure TTabListBox.FindLongestString;
var
  i: Word;
  Str: String;
begin
  FLongestString:= 0;
  for i:=0 to Items.Count -1 do
  begin
    Str:= Items[i];
    SetScrollLength(Str);
  end;
end;

procedure TTabListBox.LBDeleteString(var Msg: TMessage);
var Str: String;
begin
  if FSizeAfterdel then
  begin
    Str:= Items[Msg.wParam];
    inherited;
    if GetLBStringLength(Str) = FLongestString then
      FindLongestString;
  end
  else
    inherited;
end;              

end.
